package jp.co.isid.elportal.fileconvert.command;

import java.io.File;
import java.io.FileOutputStream;
import java.math.BigDecimal;
import java.text.DecimalFormat;

import jp.co.isid.elportal.base.cifs.exception.CifsApplicationException;
import jp.co.isid.elportal.base.config.ParameterCatalog;
import jp.co.isid.elportal.base.log.Logger;
import jp.co.isid.elportal.fileconvert.consts.CommonConstants.CifsType;
import jp.co.isid.elportal.fileconvert.util.FileUtil;

import com.itextpdf.text.Document;
import com.itextpdf.text.Element;
import com.itextpdf.text.Font;
import com.itextpdf.text.Image;
import com.itextpdf.text.PageSize;
import com.itextpdf.text.Paragraph;
import com.itextpdf.text.Rectangle;
import com.itextpdf.text.pdf.BaseFont;
import com.itextpdf.text.pdf.PdfContentByte;
import com.itextpdf.text.pdf.PdfPCell;
import com.itextpdf.text.pdf.PdfPTable;
import com.itextpdf.text.pdf.PdfWriter;

/**
 * JPEGファイルまたはPNGファイルをPdfファイルに変換するコマンドクラスです。
 *
 *
 */
public class JpgPngToPdfCommand implements Command<File, String> {

    /** フォント **/
    private static final String FONT_NAME_MINCHO = ParameterCatalog.instance().getParameter("FileConvertParam", "fontNameMincho");
    /** 最大縮小率 **/
    private static final BigDecimal MAX_PERCENTAGE = new BigDecimal("100");
    /** PDFのdpi **/
    private static final float PDF_DPI = 72f;

    /**
     * デフォルトコンストラクタ
     */
    public JpgPngToPdfCommand() {
    }

    /**
     * JPEGファイルまたはPNGファイルをPDFに変換します。<br>
     *
     * @param imageFile 変換元ファイル
     * @return 変換後のファイル
     * @throws CommandException 変換に失敗した際にthrowされる例外
     */
    public String execute(File imageFile) throws CommandException {
        //バッチから呼び出されることを想定

        Logger.debug("JpegPngToPdfCommand execute(File imageFile)start ----------------------");

        if (!imageFile.exists()) {
            Logger.error("JPEG/PINGのPDF変換に失敗しました。-> ファイルがありません：" + imageFile.getPath());
            throw new CommandException("JPEG/PINGのPDF変換に失敗しました。");
        }

        // 作成するPDFの相対パス取得
        String absolutePath = FileUtil.getAbsolutePath(imageFile.getName(), "pdf");
        String outFile = null;
        try {
            // 作成するPDFの絶対パス取得
            outFile = FileUtil.getFullPath(absolutePath, CifsType.PDF);
        } catch (CifsApplicationException e) {
            Logger.error("JPEG/PINGのPDF変換に失敗しました。-> PDFファイルパス取得に失敗しました。", e);
            throw new CommandException("JPEG/PINGのPDF変換に失敗しました。-> PDFファイルパス取得に失敗しました。", e);
        }

        Logger.debug("JpegPngToPdfCommand execute(File imageFile)end ----------------------");

        // JPEG/PNG⇒PDF
        this.execute(imageFile, outFile);

        return absolutePath;
    }

    /**
     * JPEGファイルまたはPNGファイルをPDFに変換します。<br>
     *
     * @param imageFile 変換元ファイル
     * @param outFileName 保存ファイル
     * @return 変換後のファイル
     * @throws CommandException 変換に失敗した際にthrowされる例外
     */
    public File execute(File imageFile, String outFileName) throws CommandException {

        Logger.debug("JpegPngToPdfCommand execute(File imageFile, String outFileName)start ----------------------");

        //入力JPEG/PNGファイル
        String inFile = imageFile.getAbsolutePath();

        Logger.debug("JPEG/PING File名::" + inFile + (imageFile.exists() ? " 存在する" : " 存在しない"));

        Document document = null;

        try {
            document = new Document(PageSize.A4, 50, 50, 50, 50);

            PdfWriter writer = PdfWriter.getInstance(document, new FileOutputStream(outFileName));

            document.open();

            // 画像イメージのインスタンス生成（Gif,Jpeg,Pngに対応）
            Image img = Image.getInstance(inFile);

            // 解像度取得
            int dpiX = img.getDpiX();
            int dpiY = img.getDpiY();

            // PDF変換時のイメージサイズ
            BigDecimal imgWidthpx = new BigDecimal(Float.toString(img.getPlainWidth() * PDF_DPI / dpiX));
            BigDecimal imgHeightpx = new BigDecimal(Float.toString(img.getPlainHeight() * PDF_DPI / dpiY));

            // PDFのドキュメントサイズ
            BigDecimal docWidth = new BigDecimal(Float.toString(document.getPageSize().getWidth()));
            BigDecimal docHeight = new BigDecimal(Float.toString(document.getPageSize().getHeight()));

            // マージン
            BigDecimal imgMarginHight = new BigDecimal("100");
            BigDecimal imgMarginWidth = new BigDecimal("20");
            BigDecimal imgMarginLR = imgMarginWidth.multiply(new BigDecimal("2"));
            BigDecimal imgMarginTD = imgMarginHight.multiply(new BigDecimal("2"));


            //イメージサイズとドキュメントサイズの比率を確認します。
            BigDecimal percentage = (docWidth.subtract(imgMarginLR)).divide(imgWidthpx, 4, BigDecimal.ROUND_HALF_DOWN).multiply(MAX_PERCENTAGE);
            if (percentage.compareTo((docHeight.subtract(imgMarginTD)).divide(imgHeightpx, 4, BigDecimal.ROUND_HALF_DOWN).multiply(MAX_PERCENTAGE)) > 0) {
                percentage = (docHeight.subtract(imgMarginTD)).divide(imgHeightpx, 4, BigDecimal.ROUND_HALF_DOWN).multiply(MAX_PERCENTAGE);
            }

            // PDFのドキュメントサイズより画像が小さい場合
            if (percentage.compareTo(MAX_PERCENTAGE) > 0) {
                // 100パーセントのサイズで出力（拡大しない）
                percentage = MAX_PERCENTAGE;
            }

            //5%刻みに切り下げ
            percentage = cutDown(percentage);

            // イメージサイズを変更
            img.scalePercent(percentage.floatValue() * (PDF_DPI / dpiX), percentage.floatValue() * (PDF_DPI / dpiY));

            PdfContentByte cb = writer.getDirectContent();

            // 画像をPDFへ埋め込む
            // テーブル作成
            PdfPTable imgTable = new PdfPTable(1);
            imgTable.setTotalWidth(document.getPageSize().getWidth() - imgMarginLR.floatValue());
            imgTable.setLockedWidth(true);
            // セル作成
            PdfPCell imgCell = new PdfPCell(img);
            imgCell.setHorizontalAlignment(Element.ALIGN_CENTER);
            imgCell.setPadding(0f);
            imgCell.setBorder(Rectangle.NO_BORDER);
            // テーブルにセルを追加
            imgTable.addCell(imgCell);
            // テーブルをPDFに書き出し
            imgTable.writeSelectedRows(0, -1, imgMarginWidth.floatValue(), docHeight.subtract(imgMarginHight).floatValue(), cb);

            //縮小した場合は、縮小率を埋め込みます。
            if (percentage.compareTo(MAX_PERCENTAGE) < 0) {

                // フォント設定
                BaseFont bf = BaseFont.createFont(FONT_NAME_MINCHO, BaseFont.IDENTITY_H, BaseFont.EMBEDDED);
                Font font = new Font(bf, 8f, Font.BOLD);

                // テーブル作成
                PdfPTable fTable = new PdfPTable(3);
                fTable.setWidths(new float[]{7f, 1f, 1f});
                fTable.setTotalWidth(document.getPageSize().getWidth() - UnionCommand.ARTICLE_MARGIN_LR);
                fTable.setLockedWidth(true);
                fTable.getDefaultCell().setFixedHeight(100f);
                // セル作成
                PdfPCell emptyCell = new PdfPCell();
                emptyCell.setBorder(Rectangle.NO_BORDER);
                PdfPCell fCell = new PdfPCell(new Paragraph("縮小率：" + formatReductionRatio(percentage) + "％", font));
                fCell.setPadding(0f);
                fCell.setBorder(Rectangle.NO_BORDER);
                fCell.setHorizontalAlignment(Element.ALIGN_RIGHT);
                // テーブルにセルを追加
                fTable.addCell(emptyCell); // 空のセル
                fTable.addCell(fCell);
                fTable.addCell(emptyCell); // 空のセル
                // テーブルをPDFに書き出し
                fTable.writeSelectedRows(0, -1, UnionCommand.ARTICLE_MARGIN_L, UnionCommand.REDUCATION_RATIO_Y_POSITION, cb);
            }
            document.newPage();

        } catch (Throwable e) {
            Logger.error("JPEG/PINGのPDF変換に失敗しました。->PDFファイル作成に失敗しました。", e);
            throw new CommandException("JPEG/PINGのPDF変換に失敗しました。", e);
        } finally {
            if (document != null) {
                try {
                    document.close();
                } catch (Exception e) {
                    Logger.error("PDFファイルのclose処理に失敗しました。");
                }
            }
        }

        Logger.debug("JpegPngToPdfCommand execute(File imageFile, String outFileName)end ----------------------");

        return new File(outFileName);
    }

    /**
     *
     * 縮小率を5%刻みに切り下げます。
     *
     * @param orgVal 縮小率
     * @return 5%刻みに切り下げた縮小率
     */
    private BigDecimal cutDown(BigDecimal orgVal) {
        BigDecimal result = BigDecimal.ZERO;
        BigDecimal limitVal = new BigDecimal("5");

        if (BigDecimal.ZERO.compareTo(orgVal) != 0) {
            if (limitVal.compareTo(orgVal) > 0) {
                result = BigDecimal.ONE;
            } else {
                result = orgVal.subtract(orgVal.remainder(limitVal));
            }
        }

        return result;
    }

    /**
     * 縮小率をフォーマットします。
     *
     * @param orgVal 縮小率
     * @return 00.00にフォーマットした縮小率
     */
    private String formatReductionRatio(BigDecimal orgVal) {
        String result = null;

        DecimalFormat format = new DecimalFormat("00");
        result = format.format(orgVal.floatValue());

        return result;
    }
}
