package jp.co.isid.elportal.fileconvert.command;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.text.DecimalFormat;

import jp.co.isid.elportal.base.cifs.exception.CifsApplicationException;
import jp.co.isid.elportal.base.config.ParameterCatalog;
import jp.co.isid.elportal.base.log.Logger;
import jp.co.isid.elportal.fileconvert.consts.CommonConstants.CifsType;
import jp.co.isid.elportal.fileconvert.util.FileUtil;

import com.itextpdf.text.Document;
import com.itextpdf.text.Element;
import com.itextpdf.text.Font;
import com.itextpdf.text.Image;
import com.itextpdf.text.PageSize;
import com.itextpdf.text.Paragraph;
import com.itextpdf.text.Rectangle;
import com.itextpdf.text.pdf.BaseFont;
import com.itextpdf.text.pdf.PdfContentByte;
import com.itextpdf.text.pdf.PdfPCell;
import com.itextpdf.text.pdf.PdfPTable;
import com.itextpdf.text.pdf.PdfWriter;
import com.itextpdf.text.pdf.RandomAccessFileOrArray;
import com.itextpdf.text.pdf.codec.TiffImage;

/**
 * TiffファイルをPdfファイルに変換するコマンドクラスです。
 *
 *
 */
public class TiffToPdfCommand implements Command<File, String> {

    /** フォント **/
    private static final String FONT_NAME_MINCHO = ParameterCatalog.instance().getParameter("FileConvertParam", "fontNameMincho");
    /** 縮小率 **/
    private static final BigDecimal NO_REDUCATION = new BigDecimal("100");

    /**
     * デフォルトコンストラクタ
     */
    public TiffToPdfCommand() {
    }

    /**
     * TiffファイルをPDFに変換します。<br>
     *
     * @param tiffFile 変換元ファイル
     * @return 変換後のファイル
     * @throws CommandException 変換に失敗した際にthrowされる例外
     */
    public String execute(File tiffFile) throws CommandException {
        //バッチから呼び出されることを想定

        Logger.debug("TiffToPdfCommand execute(File tiffFile)start ----------------------");

        if (!tiffFile.exists()) {
            Logger.error("TIFFのPDF変換に失敗しました。-> TIFFファイルがありません：" + tiffFile.getPath());
            throw new CommandException("TIFFのPDF変換に失敗しました。");
        }

        // 作成するPDFの相対パス取得
        String absolutePath = FileUtil.getAbsolutePath(tiffFile.getName(), "pdf");
        String outFile = null;
        try {
            // 作成するPDFの絶対パス取得
            outFile = FileUtil.getFullPath(absolutePath, CifsType.PDF);
        } catch (CifsApplicationException e) {
            Logger.error("TIFFのPDF変換に失敗しました。-> PDFファイルパス取得に失敗しました。", e);
            throw new CommandException("TIFFのPDF変換に失敗しました。-> PDFファイルパス取得に失敗しました。", e);
        }

        Logger.debug("TiffToPdfCommand execute(File tiffFile)end ----------------------");

        // TIFF⇒PDF
        this.execute(tiffFile, outFile);

        return absolutePath;
    }

    /**
     * TiffファイルをPDFに変換します。<br>
     *
     * @param tiffFile 変換元ファイル
     * @param outFileName 保存ファイル
     * @return 変換後のファイル
     * @throws CommandException 変換に失敗した際にthrowされる例外
     */
    public File execute(File tiffFile, String outFileName) throws CommandException {

        Logger.debug("TiffToPdfCommand execute(File tiffFile, String outFileName)start ----------------------");

        //入力TIFFファイル
        String inFile = tiffFile.getAbsolutePath();

        Logger.debug("tiffFile名::" + inFile + (tiffFile.exists() ? " 存在する" : " 存在しない"));

        RandomAccessFileOrArray ra = null;
        Document document = null;
        try {
            document = new Document(PageSize.A4, 50, 50, 50, 50);

            PdfWriter writer = PdfWriter.getInstance(document, new FileOutputStream(outFileName));

            document.open();
            PdfContentByte cb = writer.getDirectContent();
            ra = new RandomAccessFileOrArray(inFile);
            int comps = TiffImage.getNumberOfPages(ra);

            //イメージの縦横ピクセル
            BigDecimal imgWidthpx = null;
            BigDecimal imgHeightpx = null;
            //ＰＤＦの縦横ピクセル
            BigDecimal docWidth = null;
            BigDecimal docHeight = null;

            //イメージの周りの余白はここをかえる
            BigDecimal imgMarginT = new BigDecimal(Float.toString(100f));
            BigDecimal imgMarginTD = imgMarginT.multiply(new BigDecimal("2"));
            BigDecimal imgMarginL = new BigDecimal(Float.toString(20f));
            BigDecimal imgMarginLR = imgMarginL.multiply(new BigDecimal("2"));
            BigDecimal p100 = new BigDecimal(Float.toString(100f));

            //印刷時のPDFのdpi
            float docDpi = 72f;

            //Tiffのdpi
            float xImgDpi = 0f;
            float yImgDpi = 0f;

            //縮小率をフッターに埋め込む為のテーブルです。
            PdfPTable fTable = new PdfPTable(3);
            fTable.setWidths(new float[]{7f, 1f, 1f});
            fTable.setTotalWidth(document.getPageSize().getWidth() - UnionCommand.ARTICLE_MARGIN_LR);
            fTable.setLockedWidth(true);
            fTable.getDefaultCell().setFixedHeight(100f);

            PdfPCell cellPrintData = null;
            PdfPCell cellReductionRatio = null;
            PdfPCell cellAllRightReserve = null;
            BaseFont bf = BaseFont.createFont(FONT_NAME_MINCHO, BaseFont.IDENTITY_H, BaseFont.EMBEDDED);
            Font reductionRatioFont = new Font(bf, 8f, Font.BOLD);
            cellPrintData = new PdfPCell();
            cellPrintData.setPadding(0f);
            cellPrintData.setBorder(Rectangle.NO_BORDER);
            cellAllRightReserve = new PdfPCell();
            cellAllRightReserve.setPadding(0f);
            cellAllRightReserve.setBorder(Rectangle.NO_BORDER);


            for (int c = 0; c < comps; ++c) {
                Image img = TiffImage.getTiffImage(ra, c + 1);
                if (img == null) {
                    continue;
                }

                xImgDpi = img.getDpiX();
                yImgDpi = img.getDpiY();
                //印刷時のpixel
                imgWidthpx = new BigDecimal(Float.toString(img.getPlainWidth() * docDpi / xImgDpi));
                imgHeightpx = new BigDecimal(Float.toString(img.getPlainHeight() * docDpi / yImgDpi));

                //BigDecimal対応の前の実装.BigDecimal対応すると読みにくいので残した
                //上下左右100f は余白です。
                //float percentage = (document.getPageSize().getWidth()-200f) /  img.getWidth() * 100f;
                //if (percentage > (document.getPageSize().getHeight()-200f) / img.getHeight() * 100f) {
                //	percentage = (document.getPageSize().getHeight()-200f) / img.getHeight() * 100f;
                //}
                docWidth = new BigDecimal(Float.toString(document.getPageSize().getWidth()));
                docHeight = new BigDecimal(Float.toString(document.getPageSize().getHeight()));
                //PdfとTifの比率を確認します。
                BigDecimal percentage = (docWidth.subtract(imgMarginLR)).divide(imgWidthpx, UnionCommand.SCALE, BigDecimal.ROUND_HALF_DOWN).multiply(p100);
                if (percentage.compareTo((docHeight.subtract(imgMarginTD)).divide(imgHeightpx, UnionCommand.SCALE, BigDecimal.ROUND_HALF_DOWN).multiply(p100)) > 0) {
                    percentage = (docHeight.subtract(imgMarginTD)).divide(imgHeightpx, UnionCommand.SCALE, BigDecimal.ROUND_HALF_DOWN).multiply(p100);
                }

                //拡大はしません。
                if (percentage.compareTo(p100) > 0) {
                    percentage = p100;
                }
                //BigDecimal対応の前の実装.BigDecimal対応すると読みにくいので残した
                //if (percentage > 100) {
                //	percentage = 100f;
                //}

                //スケールを変更します。
                //標準72dpiとTifのdpiの倍率も考慮します。
                float xDpiPerAtPrint = docDpi / xImgDpi;
                float yDpiPerAtPrint = docDpi / yImgDpi;

                //5%刻みに切り下げます。
                percentage = cutDown(percentage);

                //PDF/Tifの倍率と標準72dpi/Tifの倍率をかけます。
                img.scalePercent(percentage.floatValue() * xDpiPerAtPrint, percentage.floatValue() * yDpiPerAtPrint);

                //PDFのページサイズを変更します。
                document.setPageSize(new Rectangle(docWidth.floatValue(), docHeight.floatValue()));
                //BigDecimal対応の前の実装.BigDecimal対応すると読みにくいので残した
                //document.setPageSize(new Rectangle(document.getPageSize().getWidth()
                //	, document.getPageSize().getHeight()));

                //画像をPDFへ埋め込みます。
                PdfPTable imgTable = new PdfPTable(1);
                imgTable.setTotalWidth(document.getPageSize().getWidth() - imgMarginLR.floatValue());
                imgTable.setLockedWidth(true);
                PdfPCell imgCell = new PdfPCell(img);
                imgCell.setHorizontalAlignment(Element.ALIGN_CENTER);
                imgCell.setPadding(0f);
                imgCell.setBorder(Rectangle.NO_BORDER);
                imgTable.addCell(imgCell);
                imgTable.writeSelectedRows(0, -1, imgMarginL.floatValue(), docHeight.subtract(imgMarginT).floatValue(), cb);

                //縮小した場合は、縮小率を埋め込みます。
                if (percentage.compareTo(NO_REDUCATION) < 0) {
                    fTable.addCell(cellPrintData);
                    cellReductionRatio = new PdfPCell(new Paragraph("縮小率：" + formatReductionRatio(percentage) + "％", reductionRatioFont));
                    cellReductionRatio.setPadding(0f);
                    cellReductionRatio.setBorder(Rectangle.NO_BORDER);
                    cellReductionRatio.setHorizontalAlignment(Element.ALIGN_RIGHT);
                    fTable.addCell(cellReductionRatio);
                    fTable.addCell(cellAllRightReserve);
                    fTable.writeSelectedRows(0, -1, UnionCommand.ARTICLE_MARGIN_L, UnionCommand.REDUCATION_RATIO_Y_POSITION, cb);
                }

            }
            document.newPage();

        } catch (Throwable e) {
            Logger.error("TIFFのPDF変換に失敗しました。->PDFファイル作成に失敗しました。", e);
            throw new CommandException("TIFFのPDF変換に失敗しました。", e);
        } finally {
            if (ra != null) {
                try {
                    ra.close();
                } catch (IOException e) {
                    Logger.error("RandomAccessFileOrArrayファイルのclose処理に失敗しました。");
                }
            }
            if (document != null) {
                try {
                    document.close();
                } catch (Exception e) {
                    Logger.error("PDFファイルのclose処理に失敗しました。");
                }
            }
        }

        Logger.debug("TiffToPdfCommand execute(File tiffFile, String outFileName)end ----------------------");

        return new File(outFileName);
    }

    /**
     *
     * 縮小率を5%刻みに切り下げます。
     *
     * @param orgVal 縮小率
     * @return 5%刻みに切り下げた縮小率
     */
    private BigDecimal cutDown(BigDecimal orgVal) {
        BigDecimal result = BigDecimal.ZERO;
        BigDecimal limitVal = new BigDecimal("5");

        if (BigDecimal.ZERO.compareTo(orgVal) != 0) {
            if (limitVal.compareTo(orgVal) > 0) {
                result = BigDecimal.ONE;
            } else {
                result = orgVal.subtract(orgVal.remainder(limitVal));
            }
        }

        return result;
    }

    /**
     * 縮小率をフォーマットします。
     *
     * @param orgVal 縮小率
     * @return 00.00にフォーマットした縮小率
     */
    private String formatReductionRatio(BigDecimal orgVal) {
        String result = null;

        DecimalFormat format = new DecimalFormat("00");
        result = format.format(orgVal.floatValue());

        return result;
    }
}
